SystemOrganization addCategory: #'Qwaq-PShare'!

TEmbeddableApp subclass: #QPersonalShareView
	instanceVariableNames: 'message launchSet document sharedName localFileName localModificationTime processID windowID stream serverLoop buttonState mousePoint updateSema frameNumber lastFrameSent docName allowAnyInput allowRemoteInput reasonForNoInput screenForm windowForms mouseFocus keyboardFocus zOrder colorKey cropRect'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants'
	category: 'Qwaq-PShare'!

!QPersonalShareView commentStamp: '<historical>' prior: 0!
Off-island portion of a personal share.!

!QPersonalShareView class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalShareView class methodsFor: 'class initialization' stamp: 'ar 1/26/2009 18:28'!
initialize
	"QPersonalShareView initialize"
	TEmbeddedApp registerApp: self.! !

!QPersonalShareView class methodsFor: 'accessing' stamp: 'ar 1/27/2009 16:53'!
shouldRunUponLoad

	"Personal shares will only run when clicked upon"

	^false! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 2/11/2009 12:01'!
bypassKeyboardEvent: evtBuf
	"We received an event by direct dispatch from the harness.
	Only accept it if this is the share host, otherwise resignal in the space
	so that the share host receives it via replicated signal."
	self isShareHost ifFalse:[^app future signal: #keyboardEvent with: evtBuf].
	^super bypassKeyboardEvent: evtBuf! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 2/11/2009 12:01'!
bypassMouseEvent: evtBuf
	"We received an event by direct dispatch from the harness.
	Only accept it if this is the share host, otherwise resignal in the space
	so that the share host receives it via replicated signal."
	self isShareHost ifFalse:[^app future signal: #mouseEvent with: evtBuf].
	^super bypassMouseEvent: evtBuf! !

!QPersonalShareView methodsFor: 'accessing' stamp: 'ar 6/16/2009 11:39'!
colorKey
	"If present (non-nil), replace entries with this color by transparent bits"
	^colorKey! !

!QPersonalShareView methodsFor: 'accessing' stamp: 'ar 6/16/2009 11:39'!
colorKey: aColor
	"If present (non-nil), replace entries with this color by transparent bits"
	colorKey := aColor.! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 1/29/2009 10:53'!
deferredUpdateDelay
	"How long should the host wait when the clients haven't received the last frame before trying again? This will only be used if (one or more) subscribers can't take the data as quickly as we try to push it and can consequently be much more aggressive than the updateFrameDelay."
	^50 "msecs"! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'ar 2/6/2009 01:18'!
destroy
	self isShareHost ifTrue:[QPersonalShare stopWindowShare: windowID].
	self stopServerLoop.
	super destroy.
! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'ar 1/30/2009 15:09'!
enableRemoteInput
	"Remote input initialization"

	(CMacOSXPlatform isActivePlatform and:[Smalltalk osVersion < '1050']) ifTrue:[
		allowAnyInput := false.
		allowRemoteInput := false.
		reasonForNoInput := 'Remote input is not available on Mac OSX 10.4 or earlier.'.
		^session owner 
			signalInformDialog: reasonForNoInput
			label: 'Personal Sharing Warning'.
	].

	"Do we take input from within the space at all?"
	allowAnyInput := true.
	"Do we take input from remote users?"
	allowRemoteInput := true.
! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 1/29/2009 18:53'!
findLaunchedWindow
	"Find the previously launched window.

	In theory we could look at the pid returned from launch and try to 
	use only apps with that pid. Unfortunately, the information is not 
	always reliable - various apps have a launcher process which launches 
	the actual app executable. And then there is chrome which is a bunch 
	of processes. So for now, we just ignore the pid."

	| newSet |
	newSet := QPersonalShare sharableWindows.

	"Mac heuristic: If a window with matching document name is 
	brought to the front. assume it has been activated and use it."
	CMacOSPlatform isActivePlatform ifTrue:[
		newSet first = launchSet first ifFalse:[
			((QPersonalShare windowLabelFor: newSet first) 
				includesSubstring: docName caseSensitive: false) ifTrue:["We'll take it"
					^newSet first]].
	].
	"General heuristic: Find any newly launched window..."
	newSet := newSet reject:[:id| launchSet includes: id].

	"... and if more than one, pick the one that looks most promising. 
	TODO: Look at the file name if more than one, then look at screen size."
	^newSet ifEmpty:[nil] ifNotEmpty:[newSet first].! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 2/5/2009 17:47'!
forceFullUpdate
	"Force a full update"
	screenForm := nil.! !

!QPersonalShareView methodsFor: 'accessing' stamp: 'ar 1/29/2009 11:13'!
frameNumber
	^frameNumber! !

!QPersonalShareView methodsFor: 'accessing' stamp: 'ar 1/29/2009 11:13'!
frameNumber: aNumber
	frameNumber := aNumber! !

!QPersonalShareView methodsFor: 'client loop' stamp: 'ar 1/29/2009 11:49'!
handleControl: data
	"Ignored. There is currently no control traffic downstream."! !

!QPersonalShareView methodsFor: 'client loop' stamp: 'jcg 7/28/2010 15:13'!
handleMedia: data
	"Handle incoming media"
	| rect form format imageBits |
	rect := (data networkUnsignedLongAt: 10) @ (data networkUnsignedLongAt: 14) 
			corner:	(data networkUnsignedLongAt: 18) @  (data networkUnsignedLongAt: 22).
	frameNumber := data networkUnsignedLongAt: 26.
	
	QJoshHacks receiveButIgnoreVideo ifTrue: [
		"Acknowledge receipt of frame, but throw it away
		 (no decodin' or OpenGL-uploadin')."
		app isValidFarRef ifTrue: 
			[app future receivedFrame: frameNumber from: session sessionID].
		^self
	].
	
	format := data byteAt: 30.
	imageBits := ByteArray new: data byteSize - 30.
	imageBits replaceFrom: 1 to: imageBits size with: data startingAt: 31.
	format caseOf: {
		[0]	-> [
			"Raw encoding; just use them bits"
			form := Form extent: rect extent depth: 32 bits: imageBits.
		].
		[1]	->  [
			"JPEG encoding; decompress form"
			form := Form extent: rect extent depth: 32.
			JPEGReadWriter2 new uncompress: imageBits into: form.
		].
		[2]	->  [
			"PNG encoding; decompress form"
			form := PNGReadWriter formFromStream: imageBits readStream.
		].
	} otherwise:["ignore"].
	self displayForm ifNotNil:[:d| form displayOn: d at: rect origin].
	self invalidate: rect.
	app isValidFarRef ifTrue:[
		app future receivedFrame: frameNumber from: session sessionID.
	].! !

!QPersonalShareView methodsFor: 'testing' stamp: 'ar 6/30/2009 09:11'!
hasPersonalShareColorKey
	"Answer whether personal input should be enabled."
	^session owner hasPersonalShareColorKey! !

!QPersonalShareView methodsFor: 'testing' stamp: 'ar 2/11/2009 12:18'!
hasPersonalShareInput
	"Answer whether personal input should be enabled."
	^self isShareHost and:[session owner hasPersonalShareInput]! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'ar 6/30/2009 09:10'!
initialize
	super initialize.
	allowAnyInput := false.
	allowRemoteInput := false.
	mousePoint := 0@0.
	buttonState := 0.! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'ar 2/9/2009 17:19'!
initializeEvents
	super initializeEvents.
	self runScript: #onStreamChanged: when: {app. #streamChanged}.
	self runScript: #onAlertMessage:for: when: {app. #alertMessage}.
! !

!QPersonalShareView methodsFor: 'testing' stamp: 'ar 1/27/2009 19:02'!
isShareHost
	"Am I the host for this particular share?"
	^serverLoop notNil! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'bgf 2/13/2009 19:09'!
jpegCompressionForExtent: aPt
	"Calibrate the jpg compression parameter (1-100)
	 to encode an update of the given area."
	^ 50! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'bgf 3/5/2009 20:16'!
launchDocument: aDocument
	"Part 2 of personal share launch: We have the document locally, fire up the app."
	| fileUri shortName |
	document := aDocument.
	fileUri := document get: #asLocalFileUri.
	processID := document get: #processID.
	windowID := document get: #windowID.
	shortName := ( (document get: #shortName) ifNil: [ 'No filename recorded.' ] ).

	"Try to pick up an existing share"
	(windowID == nil or:[processID == nil]) ifFalse:[
		"Find a window with same window and process id."
		QPersonalShare sharableWindows 
			detect:[:id| id = windowID "and:[(QPersonalShare windowPidFor: id) = pid]"]
			ifNone:[windowID := processID := nil]. "the window is no longer there"
	].
	(windowID == nil or:[processID == nil]) ifFalse:[
		"We have the share, run the server loop"
		^self startServerLoop.
	].

	fileUri ifNil:["This was a window share that is gone"
		^self personalShareStartFailed: 'The original window has been closed.'.
	].

	"See if the file is available"
	localFileName := FileDirectory pathFromURI: fileUri.
	(FileDirectory default fileExists: localFileName) ifFalse:[
		^self personalShareStartFailed: 'The original file is no longer present.\\' withCRs, shortName.
	].
	self updateLocalModificationTime.

	"Launch the document."
	launchSet := QPersonalShare sharableWindows.
	docName := localFileName copyAfterLast: FileDirectory pathNameDelimiter.
	processID := QPersonalShare launchFile: localFileName.
	(processID = -1) ifTrue:[
		message := 'The document application did not open.\\Click to retry...'.
		^self].
	message := 'Opening document...'.
	self startServerLoop.
! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'bgf 3/5/2009 20:18'!
launchPersonalShare
	"Part 1 of personal share launch: Grab the document and ensure it's available locally."
	| owner doc |
	owner := app get: #documentOwner.
	doc := app get: #document.
	doc ifNil:[^self personalShareStartFailed: 'The original file is no longer present.'].
	sharedName := doc get: #sharedName.
	sharedName ifNil:[
		"A local document - just launch it"
		(owner == nil or:[owner = session userLogin]) ifFalse:[
			^self personalShareStartFailed: 'This personal share belongs to ', owner, 
								String cr, 'Only that person can share it.'].
		self launchDocument: doc
	] ifNotNil:[
		"A document with server backing. Download it into the cache."
		[
			session owner downloadPShareDocument: doc andLaunch: self.
		] on: Error do: [ : ex |
			^self personalShareStartFailed: 'Could not download the document.\\' withCRs, ex description.
		]].
! !

!QPersonalShareView methodsFor: 'accessing' stamp: 'ar 1/27/2009 16:01'!
mediaController
	^session mediaController! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 2/5/2009 14:55'!
merge: r1 with: r2
	"Like Rectangle>>merge: but accounts for nil/empty rects"
	r1 ifNil:[^r2].
	r2 ifNil:[^r1].
	r1 area <= 0 ifTrue:[^r2].
	r2 area <= 0 ifTrue:[^r1].
	^r1 merge: r2! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'ar 3/4/2009 10:31'!
newFrom: appData newSession: isNewSession

	(app get: #launcherID) = (session sessionID) 
		ifTrue:[self launchPersonalShare] "launch new app"
		ifFalse:[self onStreamChanged: (app get: #streamID)]. "connect to running app"! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 2/11/2009 12:01'!
onAlertMessage: aString for: userID
	"Show an alert message for the given user"
	userID = session sessionID ifTrue:[
		session owner signal: #alertMessage with: aString
	].! !

!QPersonalShareView methodsFor: 'client loop' stamp: 'ar 6/16/2009 21:50'!
onAppExtentChanged: newExtent
	"Make the default background color to be transparent so that we avoid 
	the slight gray outline due to texture wrapping when using color keying."
	super onAppExtentChanged: newExtent.
	display form fillColor: Color transparent.! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 4/10/2009 11:50'!
onJoinComplete: userName
	QLogger logInfo: 'Join complete: ', userName.! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 4/10/2009 11:50'!
onJoinStart: userName
	QLogger logInfo: 'New user: ', userName.! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 2/11/2009 12:17'!
onKeyboardEvent: evtBuf
	"Handle keyboard input. EvtBuf is an array consisting of:
		1: Event type (keyDown, keystroke, keyUp etc)
		2: Original time stamp of event
		3: Mouse button and modifier (Alt, Ctrl, Shift) state
		4: User id
		5: Character value
	"
	| evtType buttons keyVal |
	self hasPersonalShareInput ifFalse:[^nil].
	(allowRemoteInput or:[(evtBuf at: 4) = session sessionID]) ifFalse:[^nil].
	allowAnyInput ifFalse:[^nil].

	evtType := #(keyDown keystroke keyUp) indexOf: (evtBuf at: 1).
	buttons := evtBuf at: 3.
	keyVal := evtBuf at: 5.
	"If keyboard focus is set and valid and visible, post the event there"
	(QPersonalShare isWindowVisible: keyboardFocus) 
		ifFalse:[keyboardFocus := nil].
	QPersonalShare 
		postKeyboardEvent: (keyboardFocus ifNil:[self windowFromPoint: mousePoint])
		type: evtType x: mousePoint x y: mousePoint y 
		value: keyVal buttons: buttons.
	updateSema signal.! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 6/30/2009 09:11'!
onMouseEvent: evtBuf
	"Handle mouse input. EvtBuf is an array consisting of:
		1: Event type (mouseDown, mouseMove, mouseUp etc)
		2: Original time stamp of event
		3: Mouse button and modifier (Alt, Ctrl, Shift) state
		4: User id
		5: Mouse position
	"
	| type buttons modifiers which senderID |
	(self hasPersonalShareColorKey and:[(evtBuf at: 3) allMask: 8]) 
		ifTrue:[^self trackChromaRect: evtBuf].
	self hasPersonalShareInput ifFalse:[^nil].
	senderID := evtBuf at: 4.
	(allowRemoteInput or:[senderID = session sessionID]) ifFalse:[^nil].
	allowAnyInput ifFalse:[
		"Display info for clicks on disabled pshare"
		(((evtBuf at: 3) anyMask: 7) and:[reasonForNoInput notNil])
			ifTrue:[^session owner signal: #alertMessage with: reasonForNoInput].
		^nil].

	"XXXX: Check this. We may want mouse overs for menus etc."
	((evtBuf at: 1) == #mouseOver and:[buttonState = 0]) ifTrue:[^nil].

	screenForm ifNil:[^nil]. "screen form isn't composited yet"

	mousePoint := (evtBuf at: 5) truncated + screenForm offset.
	buttons := (evtBuf at: 3) bitAnd: 7.
	modifiers := (evtBuf at: 3) bitClear: 7.

	"Always post a mouse-move to the new position first"
	type := 0.
	[QPersonalShare 
		postMouseEvent: (mouseFocus ifNil:[self windowFromPoint: mousePoint])
		type: type x: mousePoint x y: mousePoint y 
		buttons: (buttonState bitOr: modifiers).
	"Update current button state as needed"
	buttonState = buttons] whileFalse:[
		"If we encounter more than one bit being changed here only process a single one"
		which := 1 << ((buttons bitXor: buttonState) highBit - 1).
		(buttons bitAnd: which) > 0 
			ifTrue:[type := which highBit * 2 - 1]
			ifFalse:[type := which highBit * 2].

		buttonState = 0 ifTrue:[
			"With no buttons pressed, pick a new mouseFocus if necessary"
			type = 0 ifTrue:[mouseFocus := nil].
			"button down sets mouse focus" 
			(type anyMask: 1) ifTrue:[
false ifTrue:[
				"Only process mouse buttons if the window is not obscured"
				(QPersonalShare windowFromPoint: mousePoint) = 
					(self windowFromPoint: mousePoint) ifFalse:[
						^app future signal: #alertMessage 
							with: 'Cannot send input to obscured window' with: senderID
					].
].
				mouseFocus := self windowFromPoint: mousePoint.
			].
			"left button click also sets keyboard focus"
			(type = 5) ifTrue:[keyboardFocus := mouseFocus].
		].

		buttonState := buttonState bitXor: which.
	].
	"Update quickly if buttons are pressed."
	buttons = 0 ifFalse:[updateSema signal].! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 4/13/2009 16:47'!
onSpaceLeave: space
	"Stop the share when leaving the space"
	(app isValidFarRef and:[self isShareHost]) 
		ifTrue:[app future stopRunning: session userID].
! !

!QPersonalShareView methodsFor: 'client loop' stamp: 'jcg 6/16/2010 18:52'!
onStreamChanged: streamID
	"The stream ID for this share has changed.
	Subscribe to the stream and launch the client loop when completed"
	| p userName |
	streamID ifNil:[^self].
	app isValidFarRef ifFalse:[^self].
	stream ifNotNil:[stream destroy].
	userName := session owner preferences publicName asString.
	app future signal: #joinStart with: userName.
	p := self mediaController subscribe: streamID 
		in: MiramarStreamInput out: MiramarStreamOutput.
	p whenFulfilled:[:s|
		QLogger logInfo: 'Stream resolved (', s, ')'.
		(stream := s) ifNotNil:[
			self runScript: #handleMedia: when: {stream. #sendMedia}.
			self runScript: #handleControl: when: {stream. #sendControlDownstream}.
			stream play.
			app future signal: #fullUpdateNeeded.
		].
		app isValidFarRef ifTrue:[app future signal: #joinComplete with: userName].
	].! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'bgf 2/26/2009 01:46'!
personalShareStartFailed: information
	"Notify user of failure, and stop running."
	session owner signalInformDialog: information label: 'Cannot Start Personal Share'.
	app future stopRunning.! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'bgf 3/18/2009 16:31'!
publishStream
	"Publish the server-side stream."
	| pubID subID p input |
	pubID := TObjectID new.
	subID := TCryptoRandom hashMessage: pubID.
	"Publish stream, wait until connected"
	p := self mediaController publish: pubID in: MiramarStreamInput out: MiramarStreamOutput.
	p waitTimeoutSeconds: 15.
	input := p value ifNil:[^nil]. "Publish failed"
	app isValidFarRef ifFalse: [ ^ nil ].
	app future streamID: subID.
	^input! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 4/10/2009 11:50'!
runServerLoop
	"This is the loop for the share host"
	| input timeout lastKeepAlive |
	zOrder := #().
	windowForms := Dictionary new.
	updateSema := Semaphore new.
	lastFrameSent := 0.
	buttonState := 0.
	mousePoint := 0@0.

	"If we're in the midst of launching a document, try to find the document"
	"XXXX: We probably should time out after a couple of seconds"
	[windowID == nil and:[processID notNil]] whileTrue:[
		windowID := self findLaunchedWindow.
		windowID ifNil:[(Delay forSeconds: 1) wait].
	].

	app isValidFarRef ifFalse: [ ^ self ].

	"Start the window share for the given window."
	QPersonalShare startWindowShare: windowID.

	"Publish the stream"
	input := self publishStream ifNil:[
				app isValidFarRef ifTrue: [ app future stopRunning].
				^ self ].

	"We got the window id. Save it in case we need it for a restart."
	app isValidFarRef ifFalse: [ ^ self ].
	app future processID: processID.
	app future windowID: windowID.
	self runScript: #forceFullUpdate when: {app. #fullUpdateNeeded}.
	self runScript: #onJoinStart: when: {app. #joinStart}.
	self runScript: #onJoinComplete: when: {app. #joinComplete}.
	self enableRemoteInput.

	"Serve the share"
	app future startKeepAliveLoop.
	lastKeepAlive := Time millisecondClockValue.
	[windowID notNil and:[QPersonalShare isWindowValid: windowID]] whileTrue:[
		(self updateScreen: input) 
			ifTrue:[timeout := self updateFrameDelay]
			ifFalse:[timeout := self deferredUpdateDelay].
		"An update was sent out, wait for the full delay or input"
		updateSema initSignals.
		updateSema waitTimeoutMSecs: timeout.
		(Time millisecondsSince: lastKeepAlive) > 1000 ifTrue:[
			app isValidFarRef ifTrue: [
				app future keepAlive.
			].
			lastKeepAlive := Time millisecondClockValue.
		].
	].
	sharedName ifNotNil:[
		session owner 
			uploadPShareDocument: document 
			ifNotModifiedSince: localModificationTime 
			deleteWhenDone: true.
	].
	serverLoop := nil.
	self savePreview.
	app isValidFarRef ifTrue: [
		app future stopRunning.
	].
! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'bgf 3/11/2009 23:23'!
saveSharedContent
	sharedName ifNotNil:[
		session owner 
			uploadPShareDocument: document 
			ifNotModifiedSince: localModificationTime 
			deleteWhenDone: false.
		self updateLocalModificationTime.
	].
	self savePreview.
	self markClean.! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 6/17/2009 14:39'!
sendNextFrame: input updateRect: updateRect
	"Encode and send the next frame"
	| format imageBits data |
	imageBits := nil.
	(imageBits == nil and:[self useJpegEncoding]) ifTrue:[
		"Use jpeg compression"
		format := 1.
		imageBits := JPEGReadWriter2 new 
					compress: (screenForm copy: updateRect) 
					quality: (self jpegCompressionForExtent: (updateRect extent)).
	].
	(imageBits == nil and:[self usePngEncoding]) ifTrue:[
		"Use PNG encoding"
		format := 2.
		imageBits := ByteArray streamContents:[:s|
			PNGReadWriter putForm: (screenForm copy: updateRect) onStream: s.
		].
	].
	(imageBits == nil and:[self useRawEncoding]) ifTrue:[
		"Raw encoding"
		format := 0.
		imageBits := ByteArray new: (updateRect area * 4).
		(BitBlt toForm: (Form extent: updateRect extent depth: 32 bits: imageBits))
			sourceForm: screenForm;
			sourceRect: updateRect;
			combinationRule: 3;
			copyBits.
	].
	data := ByteArray new: (imageBits byteSize + 29).
	data networkUnsignedLongAt: 1 put: screenForm width.
	data networkUnsignedLongAt: 5 put: screenForm height.
	data networkUnsignedLongAt: 9 put: updateRect origin x.
	data networkUnsignedLongAt: 13 put: updateRect origin y.
	data networkUnsignedLongAt: 17 put: updateRect corner x.
	data networkUnsignedLongAt: 21 put: updateRect corner y.
	data networkUnsignedLongAt: 25 put: lastFrameSent.
	data byteAt: 29 put: format.
	data replaceFrom: 30 to: data size with: imageBits startingAt: 1.

	input sendMediaBytesDownstream: data.
! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 2/5/2009 11:48'!
shareFile: fileName
	"Share a file"
	launchSet := QPersonalShare sharableWindows asSet.
	processID := QPersonalShare launchFile: fileName.
	(processID = -1) ifTrue:[
		message := 'Document launch failed.\\Click to retry...'.
		^self].
	message := 'Launching document...'.
	self startServerLoop.! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 1/29/2009 11:00'!
startServerLoop
	"Launch the server loop for the share"
	self stopServerLoop.
	serverLoop := self fork: #runServerLoop at: Processor lowIOPriority.
! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 1/29/2009 11:00'!
stopServerLoop
	"Terminate the server loop for the share"
	serverLoop ifNotNil:[
		serverLoop terminate.
		serverLoop := nil.
	].! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 6/16/2009 17:33'!
trackChromaRect: evtBuf
	"Track the chroma keying rectangle"
	| pt |
	pt := (evtBuf at: 5) truncated.
	(evtBuf at: 1) caseOf:{
		[#mouseDown]	-> [
			((evtBuf at: 3) allMask: 12) ifTrue:[
				cropRect := pt extent: 0@0.
				colorKey := screenForm ifNotNil:[Bitmap with: (screenForm pixelValueAt: pt)].
			].
		].
		[#mouseMove]	-> [
			((evtBuf at: 3) allMask: 12) 
				ifTrue:[cropRect := cropRect encompass: pt].
		].
		[#mouseUp]	-> [
			cropRect area < 100 ifTrue:[
				cropRect := nil.
				colorKey := nil.
			].
			windowForms := Dictionary new.
			self forceFullUpdate.
		].
	} otherwise:[^self].! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 1/29/2009 10:51'!
updateFrameDelay
	"When a frame was successfully sent to all subscribers, how much should we
	wait max before we issue another update? The value here is fairly conservative."
	^500 "msecs"! !

!QPersonalShareView methodsFor: 'initialize' stamp: 'ar 2/12/2009 15:00'!
updateLocalModificationTime
	"Update the local modification time for the shared file"
	| entry |
	entry := FileDirectory default directoryEntryFor: localFileName.
	localModificationTime := entry ifNotNil:[ entry modificationTime].
! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 6/18/2010 15:20'!
updateScreen: input
	"Update my display"
	| oldestFrame updateRect screenRect oldIDs |
	app isValidFarRef ifFalse:[^false]. "get outta here"

	"Check if the subscribers have received the last frames and get out of here if not"
	oldestFrame := app get: #oldestFrameAnybodyReceived.
	(lastFrameSent - oldestFrame) < 2 ifFalse:[^false].

	"Get out here if the window is minimized"
	(QPersonalShare isWindowMinimized: windowID) ifTrue:[^true].

	"Update the main shared window itself"
	updateRect := self updateWindowID: windowID skipExtremes: false.
	screenRect := (windowForms at: windowID) offsetBox.

	"Update all the owned windows and compute damage regions"
	oldIDs := windowForms keys asSet.
	oldIDs remove: windowID ifAbsent:[].
	zOrder := OrderedCollection  new.
	QPersonalShare ownedWindowsFor: windowID do:[:ownedID|
		| oldRect newRect winRect |
		winRect := self updateWindowID: ownedID skipExtremes: true.
		winRect ifNotNil:[
			zOrder add: ownedID.
			oldIDs remove: ownedID ifAbsent:[].
			oldRect := windowForms at: ownedID ifPresent:[:f| f offsetBox].
			newRect := windowForms at: ownedID ifPresent:[:f| f offsetBox].
			oldRect = newRect ifFalse:["size / extent changed"
				updateRect := self merge: updateRect with: (self merge: oldRect with: newRect).
			].
			updateRect := self merge: updateRect with: winRect.
			screenRect := self merge: screenRect with: newRect.
		].
	].

	"Update regions for owned windows that went away"
	oldIDs do:[:ownedID|
		updateRect := self merge: updateRect with: (windowForms removeKey: ownedID) offsetBox.
	].

	"Check if the window has become invisible and if so skip the last frame
	since taking frames during a window close results in weird effects on windows."
	(QPersonalShare isWindowVisible: windowID) ifFalse:[^false].

	"Re-composite the screenForm"
	(screenForm notNil and:[screenForm extent = screenRect extent]) ifFalse:[
		"Publish the change in extent to all subscribers and give them time to catch up"
		(app future appExtent: screenRect extent) waitTimeoutSeconds: 5.

		"Need a new screenForm"
		screenForm := nil.
		screenForm := Form extent: screenRect extent depth: 32.
		screenForm fillColor: Color veryLightGray.
		updateRect := screenRect.
	].
	screenForm offset: screenRect origin.

	updateRect area <= 0 ifTrue:[^true].

	"Recomposite the screen form in the update rectangle"
	zOrder add: windowID.
	zOrder reverseDo:[:wid| | form |
		form := windowForms at: wid ifAbsent:[nil].
		(form notNil and:[updateRect intersects: form offsetBox]) ifTrue:[
			form displayOn: screenForm at: screenForm offset negated.
		].
	].

	colorKey ifNotNil:[
		"If we have a color key, replace those pixel values"
		(BitBlt toForm: screenForm)
			destRect: (updateRect translateBy: screenForm offset negated);
			"Make sure the form has alpha set"
			combinationRule: 40 "fixAlpha"; copyBits;
			"And clear out the color key"
			fillColor: colorKey; combinationRule: 39; copyBits].

	cropRect ifNotNil:[
		((updateRect translateBy: screenForm offset negated) areasOutside: cropRect) 
			do:[:r| screenForm fill: r fillColor: Color transparent].
	].

	updateRect := (updateRect intersect: screenForm offsetBox) 
		translateBy: screenForm offset negated.
	updateRect area <= 0 ifTrue:[^true].

	self sendNextFrame: input updateRect: updateRect.

	lastFrameSent := lastFrameSent + 1.

	^true! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 4/20/2010 12:48'!
updateWindowID: wid skipExtremes: aBool
	"Update the window with the given id. Return the dirty region for the window in SCREEN coordinates (i.e., offset by the window origin) so that the caller can compute bounding boxes relative to the main window if necessary."
	| form rect incr updateRect |
	form := windowForms at: wid ifAbsent:[nil].
	rect := QPersonalShare windowRectFor: wid.

	aBool ifTrue:[
		"Check for extreme window sizes. Some applications move windows
		to 10000@10000 to 'hide' them."
		(rect left < 2048 
			and:[rect right > -2048
			and:[rect top < 2048
			and:[rect bottom > -2048]]]) ifFalse:[^nil]].

	(form == nil or:[form extent ~= rect extent]) ifTrue:[
		windowForms removeKey: wid ifAbsent:[nil].
		form := Form extent: rect extent depth: 32.
		windowForms at: wid put: form.
		incr := false.
	] ifFalse:[incr := true].
	form offset: rect origin.
	updateRect := QPersonalShare windowBitmap: wid into: form incremental: incr.
	^updateRect translateBy: rect origin! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 6/16/2009 12:19'!
useJpegEncoding
	"Answer true if we should use jpeg encoding for the current frame"
	^colorKey == nil! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 6/17/2009 14:39'!
usePngEncoding
	"Answer true if we should use png encoding for the current frame"
	^self useJpegEncoding not! !

!QPersonalShareView methodsFor: 'server loop' stamp: 'ar 6/17/2009 14:40'!
useRawEncoding
	"Answer true if we should use raw encoding for the current frame"
	^false! !

!QPersonalShareView methodsFor: 'events' stamp: 'ar 2/5/2009 18:10'!
windowFromPoint: aPoint
	"Return the window at the given point, nil if none"
	zOrder reverseDo:[:wid| | form |
		form := windowForms at: wid ifAbsent:[nil].
		(form notNil and:[form offsetBox containsPoint: aPoint]) ifTrue:[^wid]].
	^nil! !

QSimpleListItem subclass: #QPersonalShareWindowListItem
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

QPersonalShareWindowListItem defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="icon" toSet="icon:" changeEvent="iconChanged">icon</field>
</fields>'!

!QPersonalShareWindowListItem class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalShareWindowListItem methodsFor: 'events' stamp: 'ar 1/29/2009 22:56'!
colorForBackground
	^Color transparent! !

!QPersonalShareWindowListItem methodsFor: 'events' stamp: 'ar 1/29/2009 22:58'!
colorForText
	^ self selected 
		ifTrue:[Color white] 
		ifFalse:[Color black].! !

!QPersonalShareWindowListItem methodsFor: 'drawing' stamp: 'bgf 7/11/2009 10:51'!
drawOn: aCanvas in: box

	super drawOn: aCanvas in: box.

	icon ifNotNil: [
		aCanvas translucentImage: icon at: (box topLeft + (4@2))
	]! !

!QPersonalShareWindowListItem methodsFor: 'accessing' stamp: 'bgf 3/8/2009 16:19'!
icon
	"Answer the icon of the receiver"
	<bewareOf: #iconChanged>
	^self propertyValueAt: #icon! !

!QPersonalShareWindowListItem methodsFor: 'accessing' stamp: 'bgf 3/8/2009 16:19'!
icon: aValue
	"Modify the receiver's icon"
	^self propertyValueAt: #icon put: aValue with: #iconChanged! !

!QPersonalShareWindowListItem methodsFor: 'events' stamp: 'ar 1/29/2009 23:01'!
onSelectedChanged
	<on: selectedChanged>
	self textColor: self colorForText.! !

!QPersonalShareWindowListItem methodsFor: 'events' stamp: 'bgf 7/11/2009 11:02'!
onValueChanged
	"Update my properties - a new value has been set."
	| txt |
	<on: valueChanged>

	value ifNil: [ ^ self ].

	"If we're going to set the text, we clear it first, then set all the attributes,
	 and install the real text only as the last step; this saves a lot of recomposing."
	txt := QPersonalShare windowLabelFor: value.
	txt = self text ifFalse: [ self text: '' ] ifTrue: [ txt := nil "see below"].
	icon := QPersonalShare windowIconFor: value.

	self textFont: QStyles listItemFont.
	self normalColor: self colorForBackground.
	self normalTextColor: Color black.
	selected ifFalse: [
		self color: self normalColor.
	].
	self textColor: self colorForText.
	self textInset: 27@0.
	self textOffset: (self offsetForText).
	"...  Style is set; compose the text if needed."
	txt ifNotNil: [ self text: txt ].
	"self balloonHelp: (value balloonString)."
	self minHeight: 20; height: 20.
! !

!QPersonalShareWindowListItem methodsFor: 'initialize' stamp: 'bgf 7/11/2009 10:50'!
setupCostume
	super setupCostume.
	self minHeight: 20; height: 20.
	self userDraw: true.
	self textEnabled: true.
	! !

CPlayer subclass: #QPersonalVNCDialog
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalVNCDialog commentStamp: '<historical>' prior: 0!
CProjectMorph open: QPersonalVNCDialog!

QPersonalVNCDialog defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="setupPage" toSet="setupPage:" changeEvent="setupPageChanged">setupPage</field>
	<field toGet="okButton" toSet="okButton:" changeEvent="okButtonChanged">okButton</field>
	<field toGet="cancelButton" toSet="cancelButton:" changeEvent="cancelButtonChanged">cancelButton</field>
	<field toGet="title" toSet="title:" changeEvent="titleChanged">title</field>
	<field toGet="hostLabel" toSet="hostLabel:" changeEvent="hostLabelChanged">hostLabel</field>
	<field toGet="hostInput" toSet="hostInput:" changeEvent="hostInputChanged">hostInput</field>
	<field toGet="portLabel" toSet="portLabel:" changeEvent="portLabelChanged">portLabel</field>
	<field toGet="portInput" toSet="portInput:" changeEvent="portInputChanged">portInput</field>
	<field toGet="passLabel" toSet="passLabel:" changeEvent="passLabelChanged">passLabel</field>
	<field toGet="passInput" toSet="passInput:" changeEvent="passInputChanged">passInput</field>
</fields>'!

!QPersonalVNCDialog class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:36'!
cancelButton
	"Answer the cancelButton of the receiver"
	<bewareOf: #cancelButtonChanged>
	^self propertyValueAt: #cancelButton! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:36'!
cancelButton: aValue
	"Modify the receiver's cancelButton"
	^self propertyValueAt: #cancelButton put: aValue with: #cancelButtonChanged! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
hostInput
	"Answer the hostInput of the receiver"
	<bewareOf: #hostInputChanged>
	^self propertyValueAt: #hostInput! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
hostInput: aValue
	"Modify the receiver's hostInput"
	^self propertyValueAt: #hostInput put: aValue with: #hostInputChanged! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
hostLabel
	"Answer the hostLabel of the receiver"
	<bewareOf: #hostLabelChanged>
	^self propertyValueAt: #hostLabel! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
hostLabel: aValue
	"Modify the receiver's hostLabel"
	^self propertyValueAt: #hostLabel put: aValue with: #hostLabelChanged! !

!QPersonalVNCDialog methodsFor: 'initialize' stamp: 'ar 2/13/2009 18:47'!
initialize
	super initialize.
	self define: #setupPage as: CPlayer new.
	self define: #cancelButton as: CPushButton new.
	self define: #okButton as: CPushButton new.

	self define: #hostLabel as: CLabel new.
	self define: #hostInput as: CInputField new.
	self define: #portLabel as: CLabel new.
	self define: #portInput as: CInputField new.
	self define: #passLabel as: CLabel new.
	self define: #passInput as: CInputField new.
! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:36'!
okButton
	"Answer the okButton of the receiver"
	<bewareOf: #okButtonChanged>
	^self propertyValueAt: #okButton! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:36'!
okButton: aValue
	"Modify the receiver's okButton"
	^self propertyValueAt: #okButton put: aValue with: #okButtonChanged! !

!QPersonalVNCDialog methodsFor: 'events' stamp: 'ar 2/13/2009 19:00'!
onCancel
	<on: fire in: cancelButton>
	value := false.
	self close.! !

!QPersonalVNCDialog methodsFor: 'events' stamp: 'ar 2/13/2009 19:07'!
onMouseDown
	<on: mouseDown>
	self onDrag.	"Be movable."! !

!QPersonalVNCDialog methodsFor: 'events' stamp: 'ar 2/13/2009 18:59'!
onOK
	<on: fire in: okButton>
	value := true.
	self close.! !

!QPersonalVNCDialog methodsFor: 'events' stamp: 'ar 2/13/2009 19:18'!
onOpen
	<on: open>
	hand keyboardFocus: passInput.! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
passInput
	"Answer the passInput of the receiver"
	<bewareOf: #passInputChanged>
	^self propertyValueAt: #passInput! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
passInput: aValue
	"Modify the receiver's passInput"
	^self propertyValueAt: #passInput put: aValue with: #passInputChanged! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
passLabel
	"Answer the passLabel of the receiver"
	<bewareOf: #passLabelChanged>
	^self propertyValueAt: #passLabel! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
passLabel: aValue
	"Modify the receiver's passLabel"
	^self propertyValueAt: #passLabel put: aValue with: #passLabelChanged! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
portInput
	"Answer the portInput of the receiver"
	<bewareOf: #portInputChanged>
	^self propertyValueAt: #portInput! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
portInput: aValue
	"Modify the receiver's portInput"
	^self propertyValueAt: #portInput put: aValue with: #portInputChanged! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
portLabel
	"Answer the portLabel of the receiver"
	<bewareOf: #portLabelChanged>
	^self propertyValueAt: #portLabel! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
portLabel: aValue
	"Modify the receiver's portLabel"
	^self propertyValueAt: #portLabel put: aValue with: #portLabelChanged! !

!QPersonalVNCDialog methodsFor: 'initialize' stamp: 'sn 3/4/2013 22:41'!
setupCostume

	super setupCostume.
	QFloatingPane setupDialog: self.
	self extent: 380@160.
	self graphicForm: (QForms find: 'PaneFill.gif'); graphicFit: #tile.


	title := CLabel on: 'VNC Connection Setup'.
	title font: QStyles controlFont;
		textEmphasis:1;
		textColor: (QToolPane titleColor);
		hResizing: #shrinkWrap;
		fullBounds.
	title topLeft: 16@11.
	self add:title.


	setupPage
			fillColor: (QColors innerPaneColor);
			fillAlpha: (0.6);
			topLeft: title bottomLeft + (-4@12);
			roundCorners: true;
			extent: (width - 24)@(height  - (setupPage top + 50));
			borderStyle: #simple; 
			borderColor: Color veryLightGray; 
			borderWidth: 1.
	self add: setupPage.

	hostLabel := CLabel on: 'Host Name:'.
	hostLabel font: QStyles controlFont;
		  hResizing: #shrinkWrap;
		  balloonHelp: (self vncHostHelp);
		  fullBounds;
		  topLeft: 10@10.
	setupPage add: hostLabel.
	hostInput topLeft: 90@10;
		hResizing: #spaceFill;
		width: (160);
		balloonHelp: (self vncHostHelp).
	setupPage add: hostInput.


	portLabel := CLabel on: 'Port:'.
	portLabel font: QStyles controlFont;
		  hResizing: #shrinkWrap;
		  balloonHelp: (self vncPortHelp);
		  fullBounds;
		  topLeft: 260@10.
	setupPage add: portLabel.

	portInput topLeft: 300@10;
		hResizing: #spaceFill;
		width: 40;
		balloonHelp: self vncPortHelp.
	setupPage add: portInput.

	passLabel := CLabel on: 'Password:'.
	passLabel font: QStyles controlFont;
		  hResizing: #shrinkWrap;
		  balloonHelp: (self vncPassHelp);
		  fullBounds;
		  topLeft: 10@40.
	setupPage add: passLabel.

	passInput topLeft: 90@40;
		hResizing: #spaceFill;
		width: 160;
		balloonHelp: self vncPassHelp.
	passInput editor font: (StrikeFont passwordFontSize: passInput textFont height).
	setupPage add: passInput.


	"DIALOG BUTTONS" 

	okButton label: 'OK';
		 font: QStyles controlFont;
	 	 extent: 70@25;
		 fullBounds;
		 topRight: 190@(self height-39);
		 roundCorners: true.
	self add: okButton.

	cancelButton label: 'Cancel'translateMe;
		 font: QStyles controlFont;
	 	 extent: 70@25;
		 fullBounds;
		 topLeft: 210@(self height - 39);
		 roundCorners: true.
	self add: cancelButton.
! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:36'!
setupPage
	"Answer the setupPage of the receiver"
	<bewareOf: #setupPageChanged>
	^self propertyValueAt: #setupPage! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:36'!
setupPage: aValue
	"Modify the receiver's setupPage"
	^self propertyValueAt: #setupPage put: aValue with: #setupPageChanged! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
title
	"Answer the title of the receiver"
	<bewareOf: #titleChanged>
	^self propertyValueAt: #title! !

!QPersonalVNCDialog methodsFor: 'accessing' stamp: 'ar 2/13/2009 18:39'!
title: aValue
	"Modify the receiver's title"
	^self propertyValueAt: #title put: aValue with: #titleChanged! !

!QPersonalVNCDialog methodsFor: 'balloons' stamp: 'ar 2/13/2009 18:46'!
vncHostHelp
	^'The name of the host running the VNC server'! !

!QPersonalVNCDialog methodsFor: 'balloons' stamp: 'ar 2/13/2009 18:46'!
vncPassHelp
	^'The password for the VNC connection'! !

!QPersonalVNCDialog methodsFor: 'balloons' stamp: 'ar 2/13/2009 18:46'!
vncPortHelp
	^'The port for the VNC server'! !

QFloatingPane subclass: #QPersonalShareWindowPane
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalShareWindowPane commentStamp: '<historical>' prior: 0!
Panel showing the available window shares. Supports drag-and-drop onto panels.!

QPersonalShareWindowPane defineFields: '<?xml  version="1.0" ?>
<fields>
	<field toGet="windowList" toSet="windowList:" changeEvent="windowListChanged">windowList</field>
</fields>'!

!QPersonalShareWindowPane class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalShareWindowPane methodsFor: 'initialize' stamp: 'ar 1/29/2009 23:15'!
initialize
	super initialize.
	self define: #windowList as: CList new.

	initialAlignment := #rightCenter.
	initialOffset := 0@0.! !

!QPersonalShareWindowPane methodsFor: 'events' stamp: 'bgf 3/5/2009 09:03'!
onDragStart
	| offset nextPos delta wid dragLabel windowLabel |
	<on: mouseDrag in: windowList>

	owner ifNil: [ ^ self ].
	owner canModifyForum ifFalse: [ ^ self ].

	wid := windowList selection ifNil:[^nil].
	(QPersonalShare isWindowValid: wid) 
		ifFalse:[^self signal: #updateWindowList].
	windowLabel := (QPersonalShare windowLabelFor:  wid) truncateWithElipsisTo: 36.

	dragLabel := QDocumentDragIcon new.
	dragLabel fileCardBacking: 'PersonalShareDocCard.png'.
	dragLabel label: windowLabel.

	dragLabel addToWorld: world.
	offset := dragLabel cursorPoint.
	hand forEach: #mouseMove do:[
		nextPos := dragLabel cursorPoint.
		delta := nextPos - offset.
		dragLabel transform: ((COffsetTransform withOffset: delta) 
			transformedBy: dragLabel transform).
	] until: #mouseUp.

	(owner notNil and: [ owner isPointInMainView: (hand position) ignoring: {dragLabel} ]) ifTrue: [
		dragLabel dragAccept.
		owner dropWindowShare: wid at: hand position
	] ifFalse: [
		dragLabel dragReject.
	].
! !

!QPersonalShareWindowPane methodsFor: 'events' stamp: 'ar 1/29/2009 22:45'!
onTick
	"TODO: Only tick when actually on-screen"
	<ticking: 0.1>
	self signal: #updateWindowList.! !

!QPersonalShareWindowPane methodsFor: 'events' stamp: 'ar 1/29/2009 22:45'!
onUpdateWindowList
	"Update the sharable windows list"
	| widList |
	<on: updateWindowList>
	widList := QPersonalShare sharableWindows.
	(widList hasEqualElements: windowList contents) 
		ifTrue: [windowList updateAll]
		ifFalse:[windowList contents: widList].! !

!QPersonalShareWindowPane methodsFor: 'initialize' stamp: 'sn 5/8/2011 21:07'!
setupCostume
	super setupCostume.

	self title: 'Sharable Windows' translateMe.
	self defaultExtent: 320@240.
	self minExtent: 240@100.
	self extent: self defaultExtent.

	windowList
		color: (Color white);
		borderColor: (Color gray alpha: 0.3);
		itemMaker: QPersonalShareWindowListItem;
		keepCursorInView: false;
		hScrollable: false; 
		vScrollable: nil. 	"On demand."

	self setupBorderedPanel: windowList.	! !

!QPersonalShareWindowPane methodsFor: 'accessing' stamp: 'ar 1/29/2009 22:29'!
windowList
	"Answer the windowList of the receiver"
	<bewareOf: #windowListChanged>
	^self propertyValueAt: #windowList! !

!QPersonalShareWindowPane methodsFor: 'accessing' stamp: 'ar 1/29/2009 22:29'!
windowList: aValue
	"Modify the receiver's windowList"
	^self propertyValueAt: #windowList put: aValue with: #windowListChanged! !

QEmbeddedApp subclass: #QPersonalShareApp
	instanceVariableNames: 'streamID aliveTimestamp frameMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalShareApp commentStamp: '<historical>' prior: 0!
On-island portion of a personal share.!

!QPersonalShareApp class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalShareApp class methodsFor: 'instance creation' stamp: 'bgf 3/5/2009 14:32'!
fileCardBacking
	^ 'PersonalShareDocCard.png'! !

!QPersonalShareApp class methodsFor: 'instance creation' stamp: 'bgf 3/13/2009 18:15'!
newForDocument: aSharedDoc extent: aRectExtent
	^(self name: 'QPersonalShareView' extent: 256@256 data: {}) 
		document: aSharedDoc;
		yourself
! !

!QPersonalShareApp class methodsFor: 'summary' stamp: 'ar 4/16/2011 17:14'!
shortName
	"protocol - default description of the kind of application"
	^ 'OpenQwaq Personal Share'! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 2/12/2009 12:13'!
asLocalFileUri
	^document ifNotNil:[document asLocalFileUri]! !

!QPersonalShareApp methodsFor: 'keepalive' stamp: 'ar 2/17/2009 10:22'!
checkOwnerAlive
	"Check if the share is still alive"
	aliveTimestamp ifNil:[^self].
	(self island time - aliveTimestamp) > TAvatarReplica dropOptionalLoadTimeout 
		ifTrue:[^self stopRunning].
	(self future: 5000) checkOwnerAlive.! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 1/29/2009 14:38'!
documentOwner
	^document owner! !

!QPersonalShareApp methodsFor: 'initialize' stamp: 'ar 1/30/2009 16:40'!
initialize
	super initialize.
	frameMap := Dictionary new.! !

!QPersonalShareApp methodsFor: 'keepalive' stamp: 'ar 1/30/2009 16:19'!
keepAlive
	"Sent by the owner to indicate that the share is healthy and running"
	aliveTimestamp := self island time.! !

!QPersonalShareApp methodsFor: 'misc' stamp: 'ar 1/30/2009 16:32'!
oldestFrameAnybodyReceived

	"Pick up the oldest frame that any of the participants in the space have received. This method is called synchronously by the participant hosting the PShare to limit update rates. 

	A disconnect for a participant (logout, network problem) may leave the oldest frame stuck which locks up the share for all participants. The workaround (not a proper fix since the whole mechanism shouldn't even be inside an island) is to keep track of when the frames were last updated and discard any participants that have not reported in in time. 

	One twinkle with this scheme is that if the PShare host has the problem it could flood everyone with updates if all of the stamps are older than the deadline. To avoid the problem we use the latest frame received within the last n seconds OR the most recent frame anyone reported.

	The only downside of this scheme is that if a client is temporarily unavailable and then comes back up it will play catchup with the missed data. While it is doing that the other participants will have to wait until the client has caught up."

	| oldestFrame deadline frameStamp newestStamp recentFrame |
	oldestFrame := recentFrame := SmallInteger maxVal.
	newestStamp := 0.
	deadline := self island time - 10000. "10secs max"
	frameMap keysAndValuesDo:[:userID :assoc|
		frameStamp := assoc value.
		frameStamp > deadline ifTrue:[oldestFrame := oldestFrame min: assoc key].
		frameStamp > newestStamp ifTrue:[recentFrame := assoc key].
	].
	oldestFrame = SmallInteger maxVal 
		ifTrue:[^recentFrame]
		ifFalse:[^oldestFrame]
! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 2/12/2009 11:20'!
processID
	^document ifNotNil:[document processID]! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 2/12/2009 11:20'!
processID: aProcessID
	^document ifNotNil:[document processID: aProcessID]! !

!QPersonalShareApp methodsFor: 'misc' stamp: 'ar 1/30/2009 16:33'!
receivedFrame: frameNumber from: userID
	"Remember that a client received a particular frame"
	frameMap at: userID put: frameNumber -> self island time.! !

!QPersonalShareApp methodsFor: 'misc' stamp: 'bgf 3/13/2009 18:00'!
shortName
	^ document
		ifNotNil: [ document kindSummary ]
		ifNil: [ 'Personal Share'].! !

!QPersonalShareApp methodsFor: 'keepalive' stamp: 'ar 1/30/2009 16:25'!
startKeepAliveLoop
	"Start the keep-alive loop"
	aliveTimestamp := self island time.
	(self future: 5000) checkOwnerAlive.! !

!QPersonalShareApp methodsFor: 'keepalive' stamp: 'ar 1/30/2009 16:25'!
stopKeepAliveLoop
	"Stop the keep-alive loop"
	aliveTimestamp := nil.! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 1/27/2009 11:19'!
streamID
	^streamID! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 1/27/2009 11:20'!
streamID: aTObjectID
	streamID := aTObjectID.
	self signal: #streamChanged with: streamID.! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 1/26/2009 18:38'!
uri
	^document ifNotNil:[document uri]! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 1/27/2009 17:26'!
uri: aString
	^document ifNotNil:[document uri: aString]! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 2/12/2009 11:18'!
windowID
	^document ifNotNil:[document windowID]! !

!QPersonalShareApp methodsFor: 'accessing' stamp: 'ar 2/12/2009 11:19'!
windowID: aWindowID
	^document ifNotNil:[document windowID: aWindowID]! !

QEmbeddedApp subclass: #QPersonalVNCApp
	instanceVariableNames: 'streamID aliveTimestamp frameMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalVNCApp commentStamp: '<historical>' prior: 0!
This part of 'personal VNC share' where one participant hosts a VNC connection into forums.!

!QPersonalVNCApp class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalVNCApp class methodsFor: 'defaults' stamp: 'bgf 3/5/2009 16:34'!
fileCardBacking
	^ 'SharedDesktopCard.png'! !

!QPersonalVNCApp class methodsFor: 'instance creation' stamp: 'bgf 3/13/2009 18:15'!
newForDocument: aSharedDoc extent: aPixelExtent
	^(self name: 'QPersonalVNCView' extent: 256@256 data: {}) 
		document: aSharedDoc;
		yourself
! !

!QPersonalVNCApp class methodsFor: 'defaults' stamp: 'ar 4/16/2011 17:14'!
shortName
	"protocol - default description of the kind of application"
	^ 'OpenQwaq VNC Share'! !

!QPersonalVNCApp methodsFor: 'keepalive' stamp: 'ar 2/17/2009 10:22'!
checkOwnerAlive
	"Check if the share is still alive"
	aliveTimestamp ifNil:[^self].
	(self island time - aliveTimestamp) > TAvatarReplica dropOptionalLoadTimeout 
		ifTrue:[^self stopRunning].
	(self future: 5000) checkOwnerAlive.! !

!QPersonalVNCApp methodsFor: 'initialize' stamp: 'ar 2/13/2009 14:08'!
initialize
	super initialize.
	frameMap := Dictionary new.! !

!QPersonalVNCApp methodsFor: 'keepalive' stamp: 'ar 2/13/2009 14:07'!
keepAlive
	"Sent by the owner to indicate that the share is healthy and running"
	aliveTimestamp := self island time.! !

!QPersonalVNCApp methodsFor: 'misc' stamp: 'ar 2/13/2009 14:08'!
oldestFrameAnybodyReceived

	"Pick up the oldest frame that any of the participants in the space have received. This method is called synchronously by the participant hosting the PShare to limit update rates. 

	A disconnect for a participant (logout, network problem) may leave the oldest frame stuck which locks up the share for all participants. The workaround (not a proper fix since the whole mechanism shouldn't even be inside an island) is to keep track of when the frames were last updated and discard any participants that have not reported in in time. 

	One twinkle with this scheme is that if the PShare host has the problem it could flood everyone with updates if all of the stamps are older than the deadline. To avoid the problem we use the latest frame received within the last n seconds OR the most recent frame anyone reported.

	The only downside of this scheme is that if a client is temporarily unavailable and then comes back up it will play catchup with the missed data. While it is doing that the other participants will have to wait until the client has caught up."

	| oldestFrame deadline frameStamp newestStamp recentFrame |
	oldestFrame := recentFrame := SmallInteger maxVal.
	newestStamp := 0.
	deadline := self island time - 10000. "10secs max"
	frameMap keysAndValuesDo:[:userID :assoc|
		frameStamp := assoc value.
		frameStamp > deadline ifTrue:[oldestFrame := oldestFrame min: assoc key].
		frameStamp > newestStamp ifTrue:[recentFrame := assoc key].
	].
	oldestFrame = SmallInteger maxVal 
		ifTrue:[^recentFrame]
		ifFalse:[^oldestFrame]
! !

!QPersonalVNCApp methodsFor: 'misc' stamp: 'ar 2/13/2009 14:08'!
receivedFrame: frameNumber from: userID
	"Remember that a client received a particular frame"
	frameMap at: userID put: frameNumber -> self island time.! !

!QPersonalVNCApp methodsFor: 'keepalive' stamp: 'ar 2/13/2009 14:07'!
startKeepAliveLoop
	"Start the keep-alive loop"
	aliveTimestamp := self island time.
	(self future: 5000) checkOwnerAlive.! !

!QPersonalVNCApp methodsFor: 'keepalive' stamp: 'ar 2/13/2009 14:07'!
stopKeepAliveLoop
	"Stop the keep-alive loop"
	aliveTimestamp := nil.! !

!QPersonalVNCApp methodsFor: 'accessing' stamp: 'ar 2/13/2009 14:08'!
streamID
	^streamID! !

!QPersonalVNCApp methodsFor: 'accessing' stamp: 'ar 2/13/2009 14:08'!
streamID: aTObjectID
	streamID := aTObjectID.
	self signal: #streamChanged with: streamID.! !

Object subclass: #QPersonalShare
	instanceVariableNames: ''
	classVariableNames: 'IconCache'
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalShare commentStamp: '<historical>' prior: 0!
A utility class holding various functions to deal with native windows for personal share.!

!QPersonalShare class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/19/2009 10:05'!
enableDebug: aBool
	"Enable debug output"
	"QPersonalShare enableDebug: true"
	<primitive: 'primitiveEnableDebug' module: 'PSharePlugin'>
	^nil! !

!QPersonalShare class methodsFor: 'class initialization' stamp: 'ar 1/30/2009 14:26'!
flushCachedIcons
	"QPersonalShare flushCachedIcons"
	IconCache := Dictionary new.! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 4/8/2009 11:06'!
iconForDocumentFile: aFilename
	"Answer the OS icon for the document with the given file name.
	We return the largest available icon here, if a particular size is required
	the client should just call iconForDocumentFile:into: directly."
	"Example:
		(QPersonalShare iconForDocumentFile: Smalltalk imageName) display.
	"
	| iconForm |
	{128@128. 64@64. 32@32} do:[:extent|
		iconForm := self iconForDocumentFile: aFilename into: (Form extent: extent depth: 32).
		iconForm ifNotNil:[^iconForm].
	].
	^nil! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 4/8/2009 11:06'!
iconForDocumentFile: aFilename into: iconForm
	"Answer the OS icon for the document with the given file name.
	Fills in the iconForm but will fail if an icon of the requested size is not available.
	Retry with various sizes (between 256x256 and 16x16) to find one you like."
	^self primitiveGetDocumentIcon: aFilename squeakToUtf8 into: iconForm! !

!QPersonalShare class methodsFor: 'class initialization' stamp: 'ar 1/30/2009 11:16'!
initialize
	"QPersonalShare initialize"
	IconCache := Dictionary new.! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/4/2009 15:14'!
isWindowMinimized: wid
	"Returns true if the window is visible"
	<primitive: 'primitiveIsWindowMinimized' module: 'PSharePlugin'>
	^false! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/23/2009 15:12'!
isWindowSharable: wid
	"Return true if the window should be shared. Generally, the platform primitive should decide which windows are considered worthy of sharing and the code calling this should adhere to the platform recommendation. However, in some situations the platform primitive has been known to get it wrong and it can be easier if the code here has the option to ignore the platforms opinion."
	<primitive: 'primitiveIsWindowSharable' module: 'PSharePlugin'>
	^false! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/23/2009 14:43'!
isWindowValid: wid
	"Returns true if the window handle is valid"
	<primitive: 'primitiveIsWindowValid' module: 'PSharePlugin'>
	^false! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/23/2009 14:43'!
isWindowVisible: wid
	"Returns true if the window is visible"
	<primitive: 'primitiveIsWindowVisible' module: 'PSharePlugin'>
	^false! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/22/2009 16:30'!
launchFile: fullPath
	"Launches a file on the local platform. Returns the pid for the process."
	^self primitiveLaunchFile: fullPath squeakToUtf8.! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/5/2009 12:37'!
ownedWindowsFor: ownerID
	"Evaluate aBlock with all of the windows owned by the given owner window ID.
	This is used to find popups etc."
	^Array streamContents:[:s| self ownedWindowsFor: ownerID do:[:wid| s nextPut: wid]].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/4/2009 15:22'!
ownedWindowsFor: ownerID do: aBlock
	"Evaluate aBlock with all of the windows owned by the given owner window ID.
	This is used to find popups etc."
	self visibleWindowsDo:[:wid|
		wid = ownerID ifTrue:[^self]. "end of enumeration"
		(self windowOwnerFor: wid) = ownerID ifTrue:[aBlock value: wid].
	].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/28/2009 16:20'!
postKeyboardEvent: wid  type: evtType x: x y: y value: keyVal buttons: buttons
	"Posts a mouse event to a window with particular id"
	<primitive: 'primitivePostKeyboardEvent' module: 'PSharePlugin'>
	^nil
! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/27/2009 19:46'!
postMouseEvent: wid type: type x: x y: y buttons: buttons
	"Posts a mouse event to a window with particular id"
	<primitive: 'primitivePostMouseEvent' module: 'PSharePlugin'>
	^nil
! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 4/8/2009 10:31'!
primitiveGetDocumentIcon: fileName into: form
	"Queries the icon associated with the given window"
	<primitive: 'primitiveGetDocumentIcon' module: 'PSharePlugin'>
	^nil! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 16:08'!
primitiveGetWindowAppName: wid
	"Queries the app name for the given window"
	<primitive: 'primitiveGetWindowAppName' module: 'PSharePlugin'>
	^nil! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 2/19/2009 14:20'!
primitiveGetWindowBitmap: id into: form incremental: aBool
	"Queries the bitmap for the given window"
	<primitive: 'primitiveGetWindowBitmap' module: 'PSharePlugin'>
	^nil! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 15:09'!
primitiveGetWindowID: index
	"Enumeration primitive. Returns the window ID for the n-th window on the platform.
	The primitive will cache window handles when invoked with index = 0 and all
	follow-on calls will use the cache. If up-to-date information is required clients must
	call with index = 0 (this is usually achieved implicitly by looping).

	The primitive will return ALL known windows - it is up to the caller to decide which
	ones should be shared by eliminating undesirable windows."
	<primitive: 'primitiveGetWindowID' module: 'PSharePlugin'>
	^-1! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 15:14'!
primitiveGetWindowIcon: id into: form
	"Queries the icon associated with the given window"
	<primitive: 'primitiveGetWindowIcon' module: 'PSharePlugin'>
	^nil! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 15:14'!
primitiveGetWindowLabel: id
	"Queries the label for the given window"
	<primitive: 'primitiveGetWindowLabel' module: 'PSharePlugin'>
	^nil! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 15:14'!
primitiveGetWindowProcessID: id
	"Queries the pid for the process associated with the given window"
	<primitive: 'primitiveGetWindowProcessID' module: 'PSharePlugin'>
	^-1! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 3/19/2009 09:43'!
primitiveGetWindowRect: windowID
	"Queries the window rectangle for the given window"
	<primitive: 'primitiveGetWindowRect' module: 'PSharePlugin'>
	^#(0 0 0 0)! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 15:15'!
primitiveLaunchFile: fullPath
	"Launches a file on the local platform. Returns the pid for the process."
	<primitive: 'primitiveLaunchFile' module: 'PSharePlugin'>
	^-1! !

!QPersonalShare class methodsFor: 'primitives' stamp: 'ar 1/23/2009 15:15'!
primitiveWindow: wid activateShare: aBool
	"Shares / unshares the given window. Returns true if shared."
	<primitive: 'primitiveShareWindow' module: 'PSharePlugin'>
	^false! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/22/2009 16:31'!
sharableWindows
	"Return all of the sharable window IDs"
	^Array streamContents:[:s| self sharableWindowsDo:[:wid| s nextPut: wid]].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 4/7/2009 15:13'!
sharableWindowsDo: aBlock
	"Evaluate aBlock with all of the sharable window IDs.
	A window is considered sharable if:
		* It is visible
		* It meets underlying platform criteria
		* It has a label (for display in menus etc.)
	All other windows will be ignored."
	| index wid |
	index := 0.
	[wid := self primitiveGetWindowID: index.
	wid == -1] whileFalse:[
		((self isWindowVisible: wid) 
			and:[self isWindowSharable: wid]
			and:[(self windowRectFor: wid) area > 0]
			and:[(self windowLabelFor: wid) notNil]
			and:[(self windowLabelFor: wid) withBlanksTrimmed notEmpty])
				ifTrue:[aBlock value: wid].
		index := index+1.
	].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/2/2009 14:03'!
startWindowShare: wid
	"Start sharing the given window. Return true, if shared. The primitive may do a variety of things, for example restoring minimized windows, sending an activate event or whatever is necessary on the platform."
	^self primitiveWindow: wid activateShare: true! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/22/2009 23:10'!
stopWindowShare: wid
	"Stop sharing the given window"
	^self primitiveWindow: wid activateShare: false! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/4/2009 15:16'!
visibleWindows
	"Return all of the sharable window IDs"
	^Array streamContents:[:s| self visibleWindowsDo:[:wid| s nextPut: wid]].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 3/19/2009 09:43'!
visibleWindowsDo: aBlock
	"Evaluate aBlock with all of the visible window IDs.
	All other windows will be ignored."
	| index wid |
	index := 0.
	[wid := self primitiveGetWindowID: index.
	wid == -1] whileFalse:[
		((self isWindowVisible: wid) 
			and:[(self windowRectFor: wid) area > 0])
				ifTrue:[aBlock value: wid].
		index := index+1.
	].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/23/2009 16:08'!
windowAppNameFor: windowID
	"Returns the label for the given window"
	^(self primitiveGetWindowAppName: windowID) ifNotNil:[:str| str utf8ToSqueak]! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/19/2009 14:21'!
windowBitmap: wid into: form incremental: aBool
	"Queries the bitmap for the given window"
	| array |
	array := self primitiveGetWindowBitmap: wid into: form incremental: aBool.
	array ifNil:[^0@0 corner: 0@0].
	^(array at: 1) @ (array at: 2) corner: (array at: 3) @ (array at: 4)! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/9/2009 16:44'!
windowFromPoint: aPoint
	"Answer the front window from the given point"
	| rect |
	self visibleWindowsDo:[:wid|
		rect := self windowRectFor: wid.
		(rect containsPoint: aPoint) ifTrue:[^wid].
	].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/30/2009 11:16'!
windowIconFor: windowID
	"Returns the icon associated with the given window"
	^IconCache at: windowID ifAbsentPut:[
		self primitiveGetWindowIcon: windowID into: (Form extent: 16@16 depth: 32)
	].! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/23/2009 14:46'!
windowLabelFor: windowID
	"Returns the label for the given window"
	^(self primitiveGetWindowLabel: windowID) ifNotNil:[:str| str utf8ToSqueak]! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 2/4/2009 15:15'!
windowOwnerFor: wid
	"Queries the wid for the owner window"
	<primitive: 'primitiveGetWindowOwner' module: 'PSharePlugin'>
	^0! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/22/2009 17:06'!
windowPidFor: windowID
	"Returns the pid for the process associated with the given window"
	^self primitiveGetWindowProcessID: windowID! !

!QPersonalShare class methodsFor: 'accessing' stamp: 'ar 1/22/2009 23:12'!
windowRectFor: wid
	"Answers the window rectangle for the share"
	^(self primitiveGetWindowRect: wid) 
		ifNotNil:[:r| (r at: 1) @ (r at: 2) corner: (r at: 3) @ (r at: 4)].! !

QAbstractDocument subclass: #QPersonalShareDocument
	instanceVariableNames: 'owner uri windowID processID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalShareDocument commentStamp: '<historical>' prior: 0!
A document residing on a particular user's machine. Only that user has access to it and can share the document via a personal share.!

!QPersonalShareDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 1/26/2009 17:44'!
asLocalFileUri
	^uri! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'bgf 7/14/2009 12:21'!
canBeCopiedToInventory
	"Until there are server-shared shares."
	^ false! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 04:22'!
canBeOpenedPrivately
	^ false	 	"Until we sort out how to distinguish who's-who down this path.
	 			 Also offering this is a bit confusing w/respect to something already personal."! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 04:14'!
canBeSavedLocally
	"Until there are server-shared shares."
	^ false! !

!QPersonalShareDocument methodsFor: 'testing' stamp: 'jcg 4/10/2009 11:17'!
canUseAsForumWelcome
	"Doesn't make sense... can't show up unless the sharer is also present."
	^false! !

!QPersonalShareDocument methodsFor: 'persist' stamp: 'ar 2/12/2009 11:21'!
constructFrom: pStream
	super constructFrom: pStream.
	owner := pStream readString: 'owner' default: owner.
	uri := pStream readString: 'uri' default: uri.
	processID := pStream readNumber: 'processID' default: processID.
	windowID := pStream readNumber: 'windowID' default: windowID.! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'bgf 2/18/2009 16:01'!
contentSummary

	^ (uri ifNotNil:[ 
			(uri copyAfterLast:$/)
		] ifNil: [ 
			'Window ',(windowID ifNotNil: [ windowID ] ifNil: [''])
		]),
		(owner ifNotNil: [ ' (', (QStyles simplerLoginName: owner), ')']  ifNil: [''])
		! !

!QPersonalShareDocument methodsFor: 'initialize' stamp: 'ar 2/12/2009 11:18'!
copyFrom: aDocument
	"The file card system relies on document copyFrom:"
	owner := aDocument owner clone.
	uri := aDocument uri clone.
	windowID := aDocument windowID.
	processID := aDocument processID.
	
	"At tail, so resource registration is done once at the end."
	super copyFrom: aDocument.
	! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'jcg 2/5/2009 10:38'!
embeddedAppClass
	"The TEmbeddedApp subclass
	 for the viewer/editor of this kind of document."
	^ QPersonalShareApp! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'jcg 2/5/2009 10:37'!
embeddedAppClassName
	"Protocol - so long as this is non-nil we can drop such cards into display sheets."
	^self embeddedAppClass name ! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 14:34'!
fileCardBacking
	^ 'PersonalShareDocCard.png'! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'hrs 4/4/2009 17:09'!
identifier
	"answer something in the same general style as other docs, but distinct"
	^self owner, '/', (self uri ifNotNilDo: [:u | u copyAfter: $/] ifNil: [windowID])
		! !

!QPersonalShareDocument methodsFor: 'initialize' stamp: 'ar 1/30/2009 14:59'!
initialize
	super initialize.
	suffix := 'wid'.! !

!QPersonalShareDocument methodsFor: 'testing' stamp: 'ar 1/26/2009 17:43'!
isKeptLocally
	^true! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'ar 1/29/2009 14:03'!
kindSummary
	^uri ifNil:['Window Share'] ifNotNil:['Personal Share']! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 1/26/2009 17:39'!
owner
	"The owner of the personal document identified by his (persistent) login name rather than the (transient) session ID so that we can save and restore the document."
	^owner! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 1/26/2009 17:39'!
owner: aString
	"The owner of the personal document identified by his (persistent) login name rather than the (transient) session ID so that we can save and restore the document."
	owner := aString.! !

!QPersonalShareDocument methodsFor: 'persist' stamp: 'ar 2/12/2009 11:19'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeString: 'owner' value: self owner default: nil.
	pStream storeString: 'uri' value: self uri default: nil.
	pStream storeNumber: 'processID' value: self processID default: nil.
	pStream storeNumber: 'windowID' value: self windowID default: nil.! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 2/12/2009 15:58'!
processID
	"The process ID of the running(!!) application. This is used to hook up a share where the panel is opened, closed, and re-opened but the locally shared app is still present."
	^processID! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 2/12/2009 15:58'!
processID: aProcessID
	"The process ID of the running(!!) application. This is used to hook up a share where the panel is opened, closed, and re-opened but the locally shared app is still present."
	processID := aProcessID! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'bgf 2/26/2009 01:59'!
shortName
	uri ifNotNil: [
		^ (uri indexOf: $/) > 0 ifTrue: [ uri copyAfterLast: $/ ] ifFalse: [ uri ]
	] ifNil: [
		^ 'Window'
	].! !

!QPersonalShareDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 01:26'!
suffix
	"The file name suffixes that match my type."
	^uri ifNotNil:[(uri copyAfterLast: $.) ifEmpty:[ suffix ]] 
		  ifNil: [ suffix ].! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 1/26/2009 17:44'!
uri
	"The URI of the document on the local file system"
	^uri! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 1/30/2009 14:58'!
uri: aString
	"The URI of the document on the local file system"
	uri := aString.
	suffix := uri copyAfterLast: $.! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 2/12/2009 15:58'!
windowID
	"The window ID of the running(!!) application. This is used to hook up a share where the panel is opened, closed, and re-opened but the locally shared app is still present."
	^windowID! !

!QPersonalShareDocument methodsFor: 'accessing' stamp: 'ar 2/12/2009 15:58'!
windowID: aWindowID
	"The window ID of the running(!!) application. This is used to hook up a share where the panel is opened, closed, and re-opened but the locally shared app is still present."
	windowID := aWindowID! !

QAbstractDocument subclass: #QPersonalVNCDocument
	instanceVariableNames: 'vncHost vncPort'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Qwaq-PShare'!

!QPersonalVNCDocument commentStamp: '<historical>' prior: 0!
This part of 'personal VNC share' where one participant hosts a VNC connection into forums.!

!QPersonalVNCDocument class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'bgf 3/1/2009 10:17'!
canBeOpenedPrivately
	"Would require implementing custom handling in QOpenDocExternalJob"
	^ false! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'bgf 3/3/2009 04:15'!
canBeSavedLocally
	"Nothing to save"
	^ false! !

!QPersonalVNCDocument methodsFor: 'persist' stamp: 'ar 2/13/2009 16:16'!
constructFrom: pStream
	super constructFrom: pStream.
	vncHost := pStream readString: 'vncHost' default: vncHost.
	vncPort := pStream readNumber: 'vncPort' default: vncPort.! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'bgf 2/18/2009 15:53'!
contentSummary
	^vncHost, ':', vncPort! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'ar 2/13/2009 14:10'!
embeddedAppClass
	"The TEmbeddedApp subclass
	 for the viewer/editor of this kind of document."
	^ QPersonalVNCApp! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'ar 2/13/2009 14:10'!
embeddedAppClassName
	"Protocol - so long as this is non-nil we can drop such cards into display sheets."
	^self embeddedAppClass name ! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 14:34'!
fileCardBacking
	^ 'PersonalShareDocCard.png'! !

!QPersonalVNCDocument methodsFor: 'initialize' stamp: 'bgf 3/5/2009 01:21'!
initialize
	super initialize.
	suffix := 'vnc'.
	vncHost := 'localhost'.
	vncPort := 5900.! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'ar 2/13/2009 14:11'!
kindSummary
	^'VNC Share'! !

!QPersonalVNCDocument methodsFor: 'persist' stamp: 'ar 2/13/2009 16:15'!
persistOn: pStream
	super persistOn: pStream.
	pStream storeString: 'vncHost' value: self vncHost default: nil.
	pStream storeNumber: 'vncPort' value: self vncPort default: nil.! !

!QPersonalVNCDocument methodsFor: 'protocol' stamp: 'bgf 3/5/2009 02:36'!
shortName
	^self  contentSummary! !

!QPersonalVNCDocument methodsFor: 'accessing' stamp: 'ar 2/13/2009 16:08'!
vncHost
	"The VNC host name"
	^vncHost! !

!QPersonalVNCDocument methodsFor: 'accessing' stamp: 'bgf 3/5/2009 02:46'!
vncHost: aString
	"The VNC host name"
	vncHost := aString.
	self signal: #shortNameChanged.! !

!QPersonalVNCDocument methodsFor: 'accessing' stamp: 'bgf 3/5/2009 02:48'!
vncHost: host port: port
	"Common setup."
	self vncHost: host.
	self vncPort: port.! !

!QPersonalVNCDocument methodsFor: 'accessing' stamp: 'ar 2/13/2009 16:08'!
vncPort
	"The VNC port"
	^vncPort! !

!QPersonalVNCDocument methodsFor: 'accessing' stamp: 'bgf 3/5/2009 02:46'!
vncPort: aNumber
	"The VNC port"
	vncPort := aNumber.	
	self signal: #shortNameChanged.! !

TEmbeddableApp subclass: #QPersonalVNCView
	instanceVariableNames: 'painter talker stream frameNumber serverLoop updateSema lastFrameSent screenExtent buttonState'
	classVariableNames: ''
	poolDictionaries: 'EventSensorConstants RFBConstants'
	category: 'Qwaq-PShare'!

!QPersonalVNCView commentStamp: '<historical>' prior: 0!
This part of 'personal VNC share' where one participant hosts a VNC connection into forums.!

!QPersonalVNCView class methodsFor: 'LICENSE' stamp: 'ar 4/20/2011 09:46'!
LICENSE
	^'Project OpenQwaq

	Copyright (c) 2005-20011, Teleplace, Inc., All Rights Reserved

	Redistributions in source code form must reproduce the above
	copyright and this condition.

	The contents of this file are subject to the GNU General Public
	License, Version 2 (the "License"); you may not use this file
	except in compliance with the License. A copy of the License is
	available at http://www.opensource.org/licenses/gpl-2.0.php.'! !

!QPersonalVNCView class methodsFor: 'class initialization' stamp: 'ar 2/13/2009 14:46'!
initialize
	"QPersonalVNCView initialize"
	TEmbeddedApp registerApp: self.! !

!QPersonalVNCView class methodsFor: 'accessing' stamp: 'ar 2/13/2009 14:45'!
shouldRunUponLoad

	"Personal shares will only run when clicked upon"

	^false! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 17:13'!
bypassKeyboardEvent: evtBuf
	"We received an event by direct dispatch from the harness.
	Only accept it if this is the share host, otherwise resignal in the space
	so that the share host receives it via replicated signal."
	self isShareHost ifFalse:[^app future signal: #keyboardEvent with: evtBuf].
	^super bypassKeyboardEvent: evtBuf! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 17:13'!
bypassMouseEvent: evtBuf
	"We received an event by direct dispatch from the harness.
	Only accept it if this is the share host, otherwise resignal in the space
	so that the share host receives it via replicated signal."
	self isShareHost ifFalse:[^app future signal: #mouseEvent with: evtBuf].
	^super bypassMouseEvent: evtBuf! !

!QPersonalVNCView methodsFor: 'initialize' stamp: 'ar 2/13/2009 19:13'!
cancelLaunch
	"cancel the launch"
	^app isValidFarRef ifTrue:[app future stopRunning].! !

!QPersonalVNCView methodsFor: 'initialize' stamp: 'ar 2/17/2009 10:47'!
connectTo: hostName port: hostPort password: password
	"Connect to a running VNC server"
	| p |
	[talker := QRFBTalker new.
	password ifNotNil:[talker password: password].
	p := talker connectTo: hostName display: 0 portOffset: hostPort.
	p waitTimeoutSeconds: 15.
	p value == true ifFalse:[^app future stopRunning].
	painter := QRFBPainter new.
	] on: Error do:[:ex| ^app future stopRunning].
	self startServerLoop.! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 14:53'!
deferredUpdateDelay
	"How long should the host wait when the clients haven't received the last frame before trying again? This will only be used if (one or more) subscribers can't take the data as quickly as we try to push it and can consequently be much more aggressive than the updateFrameDelay."
	^50 "msecs"! !

!QPersonalVNCView methodsFor: 'initialize' stamp: 'ar 2/13/2009 19:01'!
destroy
	"Destroy this VNC share"
	self stopServerLoop.
	talker ifNotNil:[talker abort].
	super destroy.
! !

!QPersonalVNCView methodsFor: 'events' stamp: 'bgf 7/10/2009 10:47'!
encodeButtons: buttonMask
	"Answer a RFB button mask equivalent to the Squeak buttonMask."
	| buttons |
	buttons := 0.

	"IAN;s original used 
			#((1 2) (2 0) (4 -2)) do:
	 
	 but all Qwaq images use 'swap mouse buttons'  so we exchange 
	 the yellow and blue mappings."

	#((1 1) (2 1) (4 -2)) do:
		[:maskShift |
		 buttons := buttons bitOr: ((buttonMask bitAnd: maskShift first) bitShift: maskShift second)].
	
	^buttons! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 18:12'!
encodeKey: keyCode
	"Answer a key sym corresponding to the given Squeak keyCode.  Note: if the Control key is down we don't encode.  This ensures that C-l (ascii 12) remains C-l (keysym 12) rather than 'page down' (MacRoman 12) which would be 'Next' (keysym #ff56)."

	"Modified for QMS (bgf): 
	 That logic doesn't really work for connections to X-VNC.
	 Instead, for A-Z (which the VM reduces to ordinal 1..26 when CTL-down),
	 we shift them to ASCII codes.  This lets the basic app control keys
	 (Ctl-S for save, etc) work in remote X applications like OpenOffice."

	| keySym |
	(buttonState anyMask: CtrlKeyBit)
		ifFalse: [keySym := KeySyms at: keyCode+1]
		ifTrue: [
			((1 to: 26) includes: keyCode) 
				ifTrue: [ keySym := 96 + keyCode]  "back to Ascii"
				ifFalse: [ keySym := keyCode]
		].

	keySym isNil ifTrue: [keySym := keyCode].
	^keySym! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 16:06'!
forceFullUpdate
	"Send a full update request"
	talker sendFullUpdateRequest.! !

!QPersonalVNCView methodsFor: 'accessing' stamp: 'ar 2/13/2009 14:49'!
frameNumber
	^frameNumber! !

!QPersonalVNCView methodsFor: 'accessing' stamp: 'ar 2/13/2009 14:49'!
frameNumber: aNumber
	frameNumber := aNumber! !

!QPersonalVNCView methodsFor: 'client' stamp: 'ar 2/13/2009 14:11'!
handleControl: data
	"Ignored. There is currently no control traffic downstream."! !

!QPersonalVNCView methodsFor: 'client' stamp: 'ar 2/13/2009 14:11'!
handleMedia: data
	"Handle incoming media"
	| rect jpegBits form |
	rect := (data networkUnsignedLongAt: 10) @ (data networkUnsignedLongAt: 14) 
			corner:	(data networkUnsignedLongAt: 18) @  (data networkUnsignedLongAt: 22).
	frameNumber := data networkUnsignedLongAt: 26.
	jpegBits := ByteArray new: (data byteSize - 30).
	jpegBits replaceFrom:1 to: (jpegBits size) with: data startingAt: 31.
	form := Form extent: rect extent depth: 32.
	JPEGReadWriter2 new uncompress: jpegBits into: form.
	form displayOn: self displayForm at: rect origin.
	self invalidate: rect.
	app future receivedFrame: frameNumber from: session sessionID.! !

!QPersonalVNCView methodsFor: 'initialize' stamp: 'ar 2/13/2009 17:20'!
initialize
	"Initialize the receiver"
	super initialize.
	lastFrameSent := 0.
	updateSema := Semaphore new.
	buttonState := 0.! !

!QPersonalVNCView methodsFor: 'initialize' stamp: 'ar 2/13/2009 16:06'!
initializeEvents
	"Initialize my events"
	super initializeEvents.
	self runScript: #onStreamChanged: when: {app. #streamChanged}.
! !

!QPersonalVNCView methodsFor: 'testing' stamp: 'ar 2/13/2009 17:13'!
isShareHost
	^serverLoop notNil! !

!QPersonalVNCView methodsFor: 'accessing' stamp: 'ar 2/13/2009 14:49'!
mediaController
	^session mediaController! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 18:11'!
modifiersForKeystroke: keyValue buttons: buttons
	"Don't apply the modifiers (CTL) that come along in a synthetic wheel event;
	 the far end has no interest, and just wants the encoded stroke."
	((buttons anyMask: 16 "ctrl-key") and: [
		(keyValue = 30) | (keyValue = 31)]) 
			ifTrue:[^buttonState]
			ifFalse:[^buttons]! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 18:00'!
mousePoint: aPoint
	"Answer a sanitised mouse point: truncated and constrained to lie within the viewer's inner bounds."

	^((aPoint max: 0@0) min: screenExtent) truncated! !

!QPersonalVNCView methodsFor: 'initialize' stamp: 'ar 3/4/2009 10:31'!
newFrom: appData newSession: isNewSession
	"Launch the VNC server if I am the requesting party"
	(app get: #launcherID) = (session sessionID) 
		ifTrue:[session owner launchPersonalVNC: self] "launch new app"
		ifFalse:[self onStreamChanged: (app get: #streamID)]. "connect to running app"! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 18:09'!
onKeyboardEvent: evtBuf
	"Handle keyboard input. EvtBuf is an array consisting of:
		1: Event type (keyDown, keystroke, keyUp etc)
		2: Original time stamp of event
		3: Mouse button and modifier (Alt, Ctrl, Shift) state
		4: User id
		5: Character value
	"
	| buttons keyVal |
	screenExtent ifNil:[^nil]. "screen form isn't composited yet"
	(evtBuf at: 1) == #keystroke ifFalse:[^nil].
	buttons := evtBuf at: 3.
	keyVal := evtBuf at: 5.

	"QMS note - on the Mac, the keypress/release keyValues are simply wrong,
	 (low-level mac keyboard codes) and do not match the keyStroke values.
	 The keyStroke values are more conventional. 
	 To ensure balanced up-downs, we do both the down and the up here.
	 Which disables hold-key timing pretty much, but that doesn't work well anyway
	 in our multishared environment."

	self
		processModifiers: (self modifiersForKeystroke: keyVal buttons: buttons);
		sendKeyEvent: (self encodeKey: keyVal) down: true;
		sendKeyEvent: (self encodeKey: keyVal) down: false.! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 17:17'!
onMouseEvent: evtBuf
	"Handle mouse input. EvtBuf is an array consisting of:
		1: Event type (mouseDown, mouseMove, mouseUp etc)
		2: Original time stamp of event
		3: Mouse button and modifier (Alt, Ctrl, Shift) state
		4: User id
		5: Mouse position
	"
	| buttons position |
	screenExtent ifNil:[^nil]. "screen form isn't composited yet"

	position := (evtBuf at: 5) truncated.
	buttons := evtBuf at: 3.

	"XXXX: Check this. We may want mouse overs for menus etc."
	((evtBuf at: 1) == #mouseOver and:[buttonState = 0]) ifTrue:[^nil].

	self
		processModifiers: buttons;
		sendPointerEvent: buttons position: position! !

!QPersonalVNCView methodsFor: 'client' stamp: 'jcg 6/16/2010 18:52'!
onStreamChanged: streamID
	"The stream ID for this share has changed.
	Subscribe to the stream and launch the client loop when completed"
	| p |
	streamID ifNil:[^self].
	stream ifNotNil:[stream destroy].
	p := self mediaController subscribe: streamID 
		in: TRFBStreamInput out: TRFBStreamOutput.
	p whenFulfilled:[:s|
		(stream := s) ifNotNil:[
			self runScript: #handleMedia: when: {stream. #sendMedia}.
			self runScript: #handleControl: when: {stream. #sendControlDownstream}.
			stream play.
			app future signal: #fullUpdateNeeded.
		].
	].! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 17:09'!
processModifiers: buttonMask
	"Check for modifier key press/release and fake the corresponding events."

	| prevState pressed released |
	prevState := buttonState.
	buttonState := buttonMask bitShift: -3.
	pressed := (prevState bitXor: -1) bitAnd: buttonState.
	released := (buttonState bitXor: -1) bitAnd: prevState.
	self sendModifiers: released down: false.
	self sendModifiers: pressed down: true! !

!QPersonalVNCView methodsFor: 'server' stamp: 'bgf 3/18/2009 16:33'!
publishStream
	"Publish the server-side stream."
	| pubID subID p input |
	pubID := TObjectID new.
	subID := TCryptoRandom hashMessage: pubID.
	"Publish stream, wait until connected"
	p := self mediaController publish: pubID in: MiramarStreamInput out: MiramarStreamOutput.
	p waitTimeoutSeconds: 15.
	input := p value ifNil:[^nil]. "Publish failed"
	app isValidFarRef ifFalse: [ ^nil ].
	app future streamID: subID.
	^input! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 16:04'!
pumpMessages
	"Pump messages between talker and painter"
	| queue msg |
	queue := talker outQueue.
	[msg := queue nextOrNil.
	msg == nil] whileFalse:[
		msg key == #initialization ifTrue:[
			painter handleInitializationMessage: msg value.
		].
		msg key == #normal ifTrue:[
			painter handleNormalRFBMessage: msg value.
		].
	].
! !

!QPersonalVNCView methodsFor: 'server' stamp: 'bgf 3/5/2009 03:14'!
runServerLoop
	"This is the loop for the share host"
	| input timeout lastKeepAlive |
	"Start the keepalive loop right away. If the wait below fails we'll just time out."
	app future startKeepAliveLoop.

	"Publish the stream"
	input := self publishStream ifNil:[^app future stopRunning].
	"We got the window id. Save it in case we need it for a restart."
	self runScript: #forceFullUpdate when: {app. #fullUpdateNeeded}.

	"Serve the share"
	lastKeepAlive := Time millisecondClockValue.
	[talker isConnected] whileTrue:[
		(self updateScreen: input) 
			ifTrue:[	timeout := self updateFrameDelay.
					"talker sendPeriodicUpdateRequest"]
			ifFalse:[	timeout := self deferredUpdateDelay].
		"An update was sent out, wait for the full delay or input"
		updateSema initSignals.
		updateSema waitTimeoutMSecs: timeout.
		(Time millisecondsSince: lastKeepAlive) > 1000 ifTrue:[
			"Careful ... might be just now getting dis/reconnected."
			app isValidFarRef ifTrue: [
				app future keepAlive.
			].
			lastKeepAlive := Time millisecondClockValue.
		].
	].
	serverLoop := nil.
	self savePreview.
	app isValidFarRef ifTrue: [
		app future stopRunning.
	].! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 18:01'!
sendKeyEvent: keyCode down: downFlag
	"Send a key event with keyCode to keySym translation."
	talker
		sendData: (RFBKeyEvent key: keyCode down: downFlag);
		sendPeriodicUpdateRequest! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 17:11'!
sendModifiers: modifiers down: downFlag
	"Send fake key press/release events for modifier keys."

	| mask |
	mask := 1.
	ModifierMap do:
		[:keySym |
		 (modifiers bitAnd: mask) ~~ 0
			ifTrue: [self sendKeyEvent: keySym down: downFlag].
		 mask := mask bitShift: 1]! !

!QPersonalVNCView methodsFor: 'events' stamp: 'ar 2/13/2009 18:00'!
sendPointerEvent: buttonMask position: aPoint
	"Send a pointer event."
	talker
		sendData: (RFBPointerEvent
						buttonMask: (self encodeButtons: buttonMask)
						position: (self mousePoint: aPoint));
		sendPeriodicUpdateRequest! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 14:20'!
startServerLoop
	"Launch the server loop for the share"
	self stopServerLoop.
	serverLoop := self fork: #runServerLoop at: Processor lowIOPriority.
! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 14:20'!
stopServerLoop
	"Terminate the server loop for the share"
	serverLoop ifNotNil:[
		serverLoop terminate.
		serverLoop := nil.
	].! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 16:04'!
updateFrameDelay
	"When a frame was successfully sent to all subscribers, how much should we
	wait max before we issue another update? The value here is fairly conservative."
	^100 "msecs"! !

!QPersonalVNCView methodsFor: 'server' stamp: 'ar 2/13/2009 19:26'!
updateScreen: input
	"Update from/to the upstream RFB server"
	| updateRect jpegBits data screenForm oldestFrame |
	app isValidFarRef ifFalse:[^false]. "get outta here"

	"Check if the subscribers have received the last frames and get out of here if not"
	oldestFrame := app get: #oldestFrameAnybodyReceived.
	(lastFrameSent - oldestFrame) < 2 ifFalse:[^false].

	self pumpMessages.

	updateRect := painter dirtyRect ifNil:[^true].
	painter dirtyRect: nil.
	screenForm := painter image.
	screenForm extent = (0@0) ifTrue:[^true].

	"Check for screen size change"
	(screenExtent notNil and:[screenExtent = screenForm extent]) ifFalse:[
		screenExtent := screenForm extent.

		"Publish the change in extent to all subscribers and give them time to catch up"
		(app future appExtent: screenExtent) waitTimeoutSeconds: 5.
		updateRect := 0@0 extent: screenExtent.
	].

	lastFrameSent := lastFrameSent + 1.
	jpegBits := JPEGReadWriter2 new compress: (screenForm copy: updateRect) quality: 50.

	data := ByteArray new: (jpegBits byteSize + 29).
	data networkUnsignedLongAt: 1 put: screenForm width.
	data networkUnsignedLongAt: 5 put: screenForm height.
	data networkUnsignedLongAt: 9 put: updateRect origin x.
	data networkUnsignedLongAt: 13 put: updateRect origin y.
	data networkUnsignedLongAt: 17 put: updateRect corner x.
	data networkUnsignedLongAt: 21 put: updateRect corner y.
	data networkUnsignedLongAt: 25 put: lastFrameSent.
	data byteAt: 29 put: ("usingJPEG" true ifTrue: [1] ifFalse:[0]).
	data replaceFrom: 30 to: data size with: jpegBits startingAt: 1.

	input sendMediaBytesDownstream: data.

	^true! !
QPersonalShareView initialize!
QPersonalShare initialize!
QPersonalVNCView initialize!
